"""A module containing tests to demonstrate that tests are working."""


class TestTest(object):
    """Container to demonstrate that tests are running."""

    def test_trivial(self):
        """Demonstrate that tests are running."""
        assert 1 == 1
