"""A top-level namespace package for IATI."""
from .codelists import Code, Codelist  # noqa: F401
from .data import Dataset  # noqa: F401
from .rulesets import Rule, Ruleset  # noqa: F401
from .rulesets import RuleAtLeastOne, RuleDateOrder, RuleDependent, RuleNoMoreThanOne, RuleRegexMatches, RuleRegexNoMatches, RuleStartsWith, RuleSum, RuleUnique  # noqa: F401
from .schemas import ActivitySchema, OrganisationSchema  # noqa: F401

__import__('pkg_resources').declare_namespace(__name__)
